from traitlets import Unicode, List, Int, observe, Dict, Any
import ipywidgets as widgets
from flutil.shape import Box


def box2json(box):
    return ({'x': box.x_min, 'y': box.y_min,
             'width': box.width, 'height': box.height}
            if isinstance(box, Box) else None)


def json2box(json):
    return (Box.from_width_height(json['width'],
                                  json['height'],
                                  top_left=(json['x'], json['y']))
            if json is not None else None)


def imbox2json(imbox):
    return ({'index': imbox['index'],
             'box': box2json(imbox['box']),
             'text': imbox['text'],
             'style': imbox['style'] if 'style' in imbox else {}}
            if imbox is not None else None)


def json2imbox(json):
    return ({'index': json['index'],
             'box': json2box(json['box']),
             'text': json['text'],
             'style': json['style'] if 'style' in json else {}}
            if json is not None and json['box'] is not None
            else None)


@widgets.register
class ImBoxWidget(widgets.DOMWidget):
    _view_name = Unicode('ImBoxView').tag(sync=True)
    _model_name = Unicode('ImBoxModel').tag(sync=True)
    _view_module = Unicode('df-imspect-widget').tag(sync=True)
    _model_module = Unicode('df-imspect-widget').tag(sync=True)
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    img = Any().tag(sync=True, to_json=lambda p, widget: str(p))
    default_style = Dict({'stroke_width': 2,
                          'stroke_style': 'red',
                          'fill_style': '#00000000',
                          'hover_fill': '#00000088',
                          'hover_stroke': 'blue',
                          'active_fill': '#ffffff22',
                          'active_stroke': 'green',
                          'font': '.75em sans-serif'}).tag(sync=True)
    boxes = List([]).tag(sync=True,
                         to_json=lambda imboxes, widget:
                         [imbox2json(imbox) for imbox in imboxes])
    width = Int(500).tag(sync=True)
    height = Int(500).tag(sync=True)
    active_box = Any().tag(sync=True,
                           to_json=lambda imbox, widget:
                           imbox2json(imbox),
                           from_json=lambda json, widget:
                           json2imbox(json))
    hover_box = Any().tag(sync=True,
                          to_json=lambda imbox, widget:
                          imbox2json(imbox),
                          from_json=lambda json, widget:
                          json2imbox(json))

    @observe('img')
    def _observe_img(self, new):
        self.active_box = None
        self.boxes = []


@widgets.register
class CropBoxWidget(widgets.DOMWidget):
    _view_name = Unicode('CropBoxView').tag(sync=True)
    _model_name = Unicode('CropBoxModel').tag(sync=True)
    _view_module = Unicode('df-imspect-widget').tag(sync=True)
    _model_module = Unicode('df-imspect-widget').tag(sync=True)
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    img = Any().tag(sync=True, to_json=lambda p, widget: str(p))
    box = Any().tag(sync=True,
                    to_json=lambda box, widget: box2json(box),
                    from_json=lambda json, widget: json2box(json))

    width = Int(150).tag(sync=True)
    height = Int(150).tag(sync=True)

    @observe('img')
    def _observe_img(self, new):
        self.box = None


@widgets.register
class DetailsWidget(widgets.DOMWidget):
    _view_name = Unicode('DetailsView').tag(sync=True)
    _model_name = Unicode('DetailsModel').tag(sync=True)
    _view_module = Unicode('df-imspect-widget').tag(sync=True)
    _model_module = Unicode('df-imspect-widget').tag(sync=True)
    _view_module_version = Unicode('^0.1.0').tag(sync=True)
    _model_module_version = Unicode('^0.1.0').tag(sync=True)

    attrs = List().tag(sync=True)
    data = Dict().tag(sync=True,
                      to_json=lambda data, widget: {k: str(v) for k, v in
                                                    data.items()})



from ipywidgets import HBox, VBox, BoundedIntText, Dropdown, Tab
from ipywidgets import HTML, Layout, Text, Button, FloatText, Play
import ipywidgets as widgets
from typing import List, Callable, Union
import pandas as pd
from .example import ImBoxWidget, CropBoxWidget, DetailsWidget
from pathlib import Path
from PIL import Image, ImageDraw, ImageFont
import logging
from skvideo.io import FFmpegWriter
import numpy as np
from tqdm import tqdm_notebook as tqdm

DEFAULT_STYLE = {
    'stroke_width': 2,
    'stroke_color': '#ff0000',
    'fill_color': '#00000000',
    'hover_fill': '#00000088',
    'hover_stroke': '#0000ff',
    'active_fill': '#ffffff22',
    'active_stroke': '#00ff00',
    'font_family': 'arial',
    'font_size': 10
}


class ImBox(VBox):
    """Widget for inspecting images that contain bounding boxes."""
    def __init__(self, df: pd.DataFrame, box_col: str = 'box',
                 img_col: str = 'image',
                 text_cols: Union[str, List[str]] = None,
                 text_fmts: Union[Callable, List[Callable]] = None,
                 style_col: str = None):
        """
        :param pd.DataFrame df: `DataFrame` with images and boxes
        :param str box_col: column in the dataframe that contains boxes
        :param str img_col: column in the dataframe that contains image paths
        :param Union[str, List[str]] text_cols: (optional) the column(s) in the
        dataframe to use for creating the text that is shown on top of a box.
        When multiple columns are give, the text will be created by a
        comma-separated list of the contents of the given columns.
        :param Unions[Callable, List[Callable]] text_fmts: (optional) a
        callable, or list of callables, that takes the corresponding value from
        the `text_cols` column(s) as an input and returns the string to print
        for that value.
        :param str style_col: the column containing a dict of style attributes.
        Available attributes are:
            - `stroke_width`: the stroke width of a box (default 2)
            - `stroke_color`: the stroke color of a box (default 'red')
            - `fill_color`: the fill color of a box (default  '#00000000')
            - `hover_fill`: the fill color of a box when it is hovered on
              (default '#00000088')
            - `hover_stroke`: the stroke color of a box when it is hovered on
              (default 'blue')
            - `active_fill`: the fill color of a box when it is clicked on
              (default '#ffffff22')
            - `active_stroke`: the stroke color of a box when it is clicked on
              (default 'green')
            - `font_family`: the font family to use for box text (default
            'arial'). NOTE: exported text will always be Arial.
            - `font_size`: the font size in points (default 10)
        """
        if text_cols is None:
            text_cols = []
        if isinstance(text_cols, str):
            text_cols = [text_cols]
        if text_fmts is None:
            text_fmts = [None]*len(text_cols)
        if isinstance(text_fmts, Callable):
            text_fmts = [text_fmts]
        self.text_cols = text_cols
        self.text_fmts = text_fmts

        df2 = df.copy()

        def row2text(row):
            txts = row[text_cols]
            return ', '.join([fmt(txt) if fmt is
                              not None else
                              str(txt)
                              for txt, fmt in zip(txts, self.text_fmts)])

        if style_col is None:
            style_col = '_dfim_style'
            df2[style_col] = [DEFAULT_STYLE]*len(df2)
        else:
            df2[style_col] = df2[style_col].apply(lambda s:
                                                  {k: s[k] if k in s
                                                   else DEFAULT_STYLE[k]
                                                   for k in DEFAULT_STYLE})

        df2['box_text'] = df2.apply(lambda row: row2text(row), axis=1)
        df2['box_dict'] = df2.apply(lambda row: dict(index=row.name,
                                                     box=row[box_col],
                                                     text=row['box_text'],
                                                     style=row[style_col])
                                    if (box_col in row.index
                                        and row[box_col] is not None)
                                    else None,
                                    axis=1)

        self.df_img = df2.groupby(img_col).agg(list).reset_index()
        self.df = df
        self.img_col = img_col
        self.box_col = box_col

        # SELECTION widget
        self.idx_wgt = BoundedIntText(value=0,
                                      min=0,
                                      max=len(self.df_img) - 1,
                                      step=1,
                                      description='Choose index',
                                      disabled=False)
        self.drop_wgt = Dropdown(options=self.df_img[img_col],
                                 description='or image',
                                 value=None,
                                 disabled=False)
        self.drop_wgt.observe(self.drop_changed, names='value')
        self.idx_wgt.observe(self.idx_changed, names='value')
        self.imsel_wgt = VBox([self.idx_wgt, self.drop_wgt])
        self.imsel_wgt.layout = Layout(margin='auto')

        # IMAGE PANE
        self.img_title = HTML(placeholder='(Image path)')
        self.img_title.layout = Layout(margin='auto')
        self.imbox_wgt = ImBoxWidget()
        self.imbox_wgt.layout = Layout(margin='1em auto')
        self.imbox_wgt.observe(self.box_changed, names='active_box')
        self.imbox_wgt.observe(self.img_changed, names='img')

        # DETAILS PANE
        self.crop_wgt = CropBoxWidget()
        self.crop_wgt.layout = Layout(margin='0 1em')
        self.detail_wgt = DetailsWidget()
        self.detail_wgt.layout = Layout(margin='auto')
        self.detail_pane = HBox([self.crop_wgt, self.detail_wgt])
        self.detail_pane.layout = Layout(margin='1em auto')

        # PLAY widget
        self.play_btns = Play(interval=100,
                              value=0,
                              min=0,
                              max=len(self.df_img) - 1,
                              step=1,
                              description="Play",
                              disabled=False)
        self.play_slider = widgets.IntSlider(value=0,
                                             min=0,
                                             max=len(self.df_img) - 1,
                                             step=1)
        widgets.jslink((self.play_btns, 'value'), (self.idx_wgt, 'value'))
        widgets.jslink((self.play_btns, 'value'), (self.play_slider, 'value'))

        self.play_wgt = widgets.HBox([self.play_btns,
                                      self.play_slider])
        self.play_wgt.layout = Layout(margin='auto')

        # IMAGE EXPORT widget
        self.imexp_dest = Text(description='Output file',
                               value='output/output.png')
        self.imexp_btn = Button(description='Export')
        self.imexp_btn.on_click(self.export_img)
        self.imexp_wgt = HBox([self.imexp_dest, self.imexp_btn])

        # VIDEO EXPORT widget
        self.videxp_dest = Text(description='Output file',
                                value='output/output.mp4')
        self.videxp_start = BoundedIntText(value=0,
                                           min=0,
                                           max=len(self.df_img) - 1,
                                           step=1,
                                           description='From index',
                                           disabled=False)
        self.videxp_start.observe(self.vididx_changed, names='value')
        self.videxp_end = BoundedIntText(value=0,
                                         min=0,
                                         max=len(self.df_img) - 1,
                                         step=1,
                                         description='Until index',
                                         disabled=False)
        self.videxp_end.observe(self.vididx_changed, names='value')
        self.videxp_fps = FloatText(value=30, description='FPS')
        self.videxp_btn = Button(description='Export')
        self.videxp_btn.on_click(self.export_vid)

        self.videxp_wgt = VBox([HBox([self.videxp_start, self.videxp_end]),
                                HBox([self.videxp_dest, self.videxp_fps]),
                                self.videxp_btn])
        self.exp_wgt = Tab(children=[self.imexp_wgt, self.videxp_wgt])
        self.exp_wgt.set_title(0, 'Export image')
        self.exp_wgt.set_title(1, 'Export video')
        self.exp_wgt.layout = Layout(margin='0 1em')

        super().__init__([self.imsel_wgt,
                          VBox([self.img_title,
                                self.imbox_wgt,
                                self.play_wgt,
                                self.detail_pane]),
                          self.exp_wgt])
        self.idx_changed({'new': 0})

    def box_changed(self, change):
        if change['new'] is None:
            self.detail_wgt.data = {}
            self.crop_wgt.box = None
        else:
            new_idx = change['new']['index']
            self.detail_wgt.data = dict(self.df.loc[new_idx])
            self.crop_wgt.box = change['new']['box']

    def img_changed(self, change):
        new_img = change['new']
        self.detail_wgt.data = {}
        self.crop_wgt.img = new_img
        self.img_title.value = f'Image path: <a href="{new_img}">{new_img}</a>'
        self.imexp_dest.value = f'output/{Path(new_img).stem}.png'
        self.imexp_btn.button_style = ''
        self.imexp_btn.description = 'Export'
        self.imexp_btn.disabled = False

    def drop_changed(self, change):
        idx = self.df_img[self.df_img[self.img_col] == change['new']].index[0]
        self.idx = idx
        self.imbox_wgt.img = self.df_img.loc[idx, self.img_col]
        self.imbox_wgt.boxes = self.df_img.loc[idx, 'box_dict']
        self.idx_wgt.value = idx

    def idx_changed(self, change):
        self.idx = change['new']
        self.imbox_wgt.img = self.df_img.loc[self.idx, self.img_col]
        self.imbox_wgt.boxes = self.df_img.loc[self.idx, 'box_dict']
        self.drop_wgt.value = self.imbox_wgt.img

    def vididx_changed(self, change):
        start = self.videxp_start.value
        end = self.videxp_end.value
        self.videxp_dest.value = f'output/{start}_{end}.mp4'
        self.videxp_btn.button_style = ''
        self.videxp_btn.description = 'Export'
        self.videxp_btn.disabled = False

    def get_pilim_from_idx(self, idx):
        """Return the processed PIL image that belongs to an image index.
        """
        im = Image.open(self.df_img.loc[idx, self.img_col])
        draw = ImageDraw.Draw(im, mode='RGBA')

        box_dicts = self.df_img.loc[idx, 'box_dict']
        for bd in box_dicts:
            box = bd['box']
            draw.rectangle(list(box)[:4],
                           fill=bd['style']['fill_color'],
                           outline=bd['style']['stroke_color'],
                           width=bd['style']['stroke_width'])

            fontfile = str(Path(__file__).parent / 'etc/Arial.ttf')

            # size*4 to make it look more similar to example in widget
            fontsize = bd['style']['font_size']*4
            font = ImageFont.truetype(fontfile,
                                      size=fontsize)
            w, h = draw.textsize(bd['text'], font=font)
            draw.text((box.x_min, box.y_min - h),
                      text=bd['text'], fill=bd['style']['stroke_color'],
                      font=font)
        return im

    def export_img(self, button):
        self.imexp_btn.disabled = True
        self.imexp_btn.description = 'Exporting...'
        im = self.get_pilim_from_idx(self.idx)
        try:
            im.save(self.imexp_dest.value)
            self.imexp_btn.button_style = 'success'
            self.imexp_btn.description = 'Export Successful'
        except (IOError, KeyError) as e:
            self.imexp_btn.button_style = 'danger'
            self.imexp_btn.description = 'Export Failed'
            logging.exception('Export Failed')

    def export_vid(self, button):
        self.videxp_btn.disabled = True
        self.videxp_btn.description = 'Exporting...'
        fps = str(self.videxp_fps.value)
        writer = FFmpegWriter(self.videxp_dest.value,
                              inputdict={'-framerate': fps})

        for idx in tqdm(range(self.videxp_start.value,
                              self.videxp_end.value)):
            im = self.get_pilim_from_idx(idx)
            writer.writeFrame(np.array(im))

        try:
            writer.close()
            self.videxp_btn.button_style = 'success'
            self.videxp_btn.description = 'Export successful'
        except OSError as e:
            self.videxp_btn.button_style = 'danger'
            self.videxp_btn.description = 'Export failed'
            logging.exception('Export Failed')



from __future__ import print_function
from setuptools import setup, find_packages, Command
from setuptools.command.sdist import sdist
from setuptools.command.build_py import build_py
from setuptools.command.egg_info import egg_info
from subprocess import check_call
import os
import sys
import platform

here = os.path.dirname(os.path.abspath(__file__))
node_root = os.path.join(here, 'js')
is_repo = os.path.exists(os.path.join(here, '.git'))

npm_path = os.pathsep.join([
    os.path.join(node_root, 'node_modules', '.bin'),
                os.environ.get('PATH', os.defpath),
])

from distutils import log
log.set_verbosity(log.DEBUG)
log.info('setup.py entered')
log.info('$PATH=%s' % os.environ['PATH'])

LONG_DESCRIPTION = 'Pandas DataFrame inspector for images with boxes'

def js_prerelease(command, strict=False):
    """decorator for building minified js/css prior to another command"""
    class DecoratedCommand(command):
        def run(self):
            jsdeps = self.distribution.get_command_obj('jsdeps')
            if not is_repo and all(os.path.exists(t) for t in jsdeps.targets):
                # sdist, nothing to do
                command.run(self)
                return

            try:
                self.distribution.run_command('jsdeps')
            except Exception as e:
                missing = [t for t in jsdeps.targets if not os.path.exists(t)]
                if strict or missing:
                    log.warn('rebuilding js and css failed')
                    if missing:
                        log.error('missing files: %s' % missing)
                    raise e
                else:
                    log.warn('rebuilding js and css failed (not a problem)')
                    log.warn(str(e))
            command.run(self)
            update_package_data(self.distribution)
    return DecoratedCommand

def update_package_data(distribution):
    """update package_data to catch changes during setup"""
    build_py = distribution.get_command_obj('build_py')
    # distribution.package_data = find_package_data()
    # re-init build_py options which load package_data
    build_py.finalize_options()


class NPM(Command):
    description = 'install package.json dependencies using npm'

    user_options = []

    node_modules = os.path.join(node_root, 'node_modules')

    targets = [
        os.path.join(here, 'dfimspect', 'static', 'extension.js'),
        os.path.join(here, 'dfimspect', 'static', 'index.js')
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def get_npm_name(self):
        npmName = 'npm';
        if platform.system() == 'Windows':
            npmName = 'npm.cmd';
            
        return npmName;
    
    def has_npm(self):
        npmName = self.get_npm_name();
        try:
            check_call([npmName, '--version'])
            return True
        except:
            return False

    def should_run_npm_install(self):
        package_json = os.path.join(node_root, 'package.json')
        node_modules_exists = os.path.exists(self.node_modules)
        return self.has_npm()

    def run(self):
        has_npm = self.has_npm()
        if not has_npm:
            log.error("`npm` unavailable.  If you're running this command using sudo, make sure `npm` is available to sudo")

        env = os.environ.copy()
        env['PATH'] = npm_path

        if self.should_run_npm_install():
            log.info("Installing build dependencies with npm.  This may take a while...")
            npmName = self.get_npm_name();
            check_call([npmName, 'install'], cwd=node_root, stdout=sys.stdout, stderr=sys.stderr)
            os.utime(self.node_modules, None)

        for t in self.targets:
            if not os.path.exists(t):
                msg = 'Missing file: %s' % t
                if not has_npm:
                    msg += '\nnpm is required to build a development version of a widget extension'
                raise ValueError(msg)

        # update package data in case this created new files
        update_package_data(self.distribution)

version_ns = {}
with open(os.path.join(here, 'dfimspect', '_version.py')) as f:
    exec(f.read(), {}, version_ns)

setup_args = {
    'name': 'dfimspect',
    'version': version_ns['__version__'],
    'description': 'Pandas DataFrame inspector for images with boxes',
    'long_description': LONG_DESCRIPTION,
    'include_package_data': True,
    'data_files': [
        ('share/jupyter/nbextensions/df-imspect-widget', [
            'dfimspect/static/extension.js',
            'dfimspect/static/index.js',
            'dfimspect/static/index.js.map',
        ],),
        ('etc/jupyter/nbconfig/notebook.d' ,['df-imspect-widget.json'])
    ],
    'install_requires': [
        'ipywidgets>=7.0.0',
        'pandas',
        'matplotlib',
        'Pillow',
        'flutil',
    ],
    'packages': find_packages(),
    'zip_safe': False,
    'cmdclass': {
        'build_py': js_prerelease(build_py),
        'egg_info': js_prerelease(egg_info),
        'sdist': js_prerelease(sdist, strict=True),
        'jsdeps': NPM,
    },

    'author': 'Floris De Feyter',
    'author_email': 'floris.defeyter@kuleuven.be',
    'url': 'https://github.com//df-imspect-widget',
    'keywords': [
        'ipython',
        'jupyter',
        'widgets',
    ],
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Framework :: IPython',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Multimedia :: Graphics',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
}

setup(**setup_args)
